//----------------------------------------------------------------------------------------
// error_computation.cpp: Contains functions computing norm of error or change for unknowns
//----------------------------------------------------------------------------------------

#include <cmath>
#include <iostream>
#include "error_computation.H"
#include "mesh_container.H"

extern MESH mesh;

// Global error computation in 3D
// Parameters: t - time; func - analytical solution; norm_type - norm choice index;
//             error - error in the given norm
void global_error(real t, void func(real, const real[d_space], real[]), int norm_type, real error[])
{
  for (int i=0; i<d_var_3d; i++)
    PRISM_SW::norm(i)=0.;
  switch (norm_type)
  {
  case 0: // L^inf norm
    break;
  case 1: // L^2 norm
    {
      IterateOverList(t, mesh.prism_list, func, &PRISM_SW::Global_error_L2);
      for (int i=0; i<d_var_3d; i++)
        error[i]=std::sqrt(PRISM_SW::norm(i));
      break;
    }
  case 2: // L^1 norm
    break;
  default: std::cerr << "File error_computation.cpp: Unknown norm type!\n";
    std::exit(1);
  }
}

// Global change between two consecutive time steps in 3D
// Parameters: dt - time step; norm_type - norm choice index; change - change in the given norm
void global_change(real dt, int norm_type, real change[])
{
  for (int i=0; i<d_var_3d_dt; i++)
    PRISM_SW::norm(i)=0.;
  switch (norm_type)
  {
  case 0: // L^inf norm
    break;
  case 1: // L^2 norm
    {
      IterateOverList(mesh.prism_list, &PRISM_SW::Global_change_L2);
      for (int i=0; i<d_var_3d_dt; i++)
        change[i]=dt*std::sqrt(PRISM_SW::norm(i));
      break;
    }
  case 2: // L^1 norm
    break;
  default: std::cerr << "File error_computation.cpp: Unknown norm type!\n";
    std::exit(1);
  }
}

// Error in surface elevation computation
// Parameters: t - time; func - analytical solution; norm_type - norm choice index;
// Return value: error in the given norm
real elevation_error(real t, void func(real, const real[d_space_2d], real[]), int norm_type)
{
  long_real result=0.;
  TRIANGLE_2D_SW::norm()=0.;

  switch (norm_type)
  {
  case 0: // L^inf norm
    break;
  case 1: // L^2 norm
    {
      IterateOverList(t, mesh.triangle_2d_list, func, &TRIANGLE_2D_SW::Elevation_error_L2);
      result=std::sqrt(TRIANGLE_2D_SW::norm());
      break;
    }
  case 2: // L^1 norm
    break;
  default: std::cerr << "File error_computation.cpp: Unknown norm type!\n";
    std::exit(1);
  }
  return result;
}

// Change in surface elevation between two consecutive time steps in 3D
// Parameters: dt - time step; norm_type - norm choice index;
// Return value: change in the given norm
real elevation_change(real dt, int norm_type)
{
  long_real result=0.;
  TRIANGLE_2D_SW::norm()=0.;
  switch (norm_type)
  {
  case 0: // L^inf norm
    break;
  case 1: // L^2 norm
    {
      IterateOverList(mesh.triangle_2d_list, &TRIANGLE_2D_SW::Elevation_change_L2);
      result=dt*std::sqrt(TRIANGLE_2D_SW::norm());
      break;
    }
  case 2: // L^1 norm
    break;
  default: std::cerr << "File error_computation.cpp: Unknown norm type!\n";
    std::exit(1);
  }
  return result;
}

#ifdef MASS_CONSERVATION_KEY
// Tracer mass conservation error
// Parameters: t - time
// Return value: error in mass balance
long_real tracer_mass_error(real t)
{
  long_real mass = std::abs(PRISM_SW::mass())<1.e-10 ? 1. : PRISM_SW::mass();

  PRISM_SW::norm(VAR_3D_DT_INDEX_INV[TRACER])=0.;
  IterateOverList(mesh.prism_list, &PRISM_SW::Tracer_Mass);
  return (PRISM_SW::norm(VAR_3D_DT_INDEX_INV[TRACER])-PRISM_SW::mass())/mass;
}  
#endif
