#include <cmath>
#include "map_info.H"

//--------------------------------------------------------------------------------------
// MapEdge Class: Map from [0, 1] to an edge in 2D
//--------------------------------------------------------------------------------------

// Compute affine mapping coefficients from reference to physical space
// Parameters: coord - edge node coordinates
void MapEdge::UpdateMap(const real* coord[d_space]) 
{
  ref_point[0]=coord[0][0];
  ref_point[1]=coord[0][1];
  map[0]=coord[1][0]-coord[0][0]; 
  map[1]=coord[1][1]-coord[0][1];
  length=std::sqrt(map[0]*map[0]+map[1]*map[1]);
  ind = std::abs(map[0])>std::abs(map[1]) ? 0 : 1;
}


//--------------------------------------------------------------------------------------
// MapTriangle Class: Map from {(0, 0), (0, 1), (1, 0)} to a triangle in 2D
//--------------------------------------------------------------------------------------

// Compute affine mapping coefficients from reference to physical space
// Parameters: coord - triangle node coordinates
void MapTriangle::UpdateMap(const real* coord[d_space]) 
{
  for (int i=0; i<d_space_2d; ref_point[i]=coord[0][i], ref_point_inv[i]=0., i++)
    for (int j=0; j<d_space_2d; j++)
      map[i][j]=coord[j+1][i]-coord[0][i];
  scale=std::abs(map[0][0]*map[1][1]-map[0][1]*map[1][0]);
  diam=std::sqrt(scale);
  map_inv[0][0]=map[1][1]/scale;
  map_inv[0][1]=-map[0][1]/scale;
  map_inv[1][0]=-map[1][0]/scale;
  map_inv[1][1]=map[0][0]/scale;
  for (int i=0; i<d_space_2d; i++)
    for (int j=0; j<d_space_2d; j++)
      ref_point_inv[i]-=map_inv[i][j]*ref_point[j];
}


//--------------------------------------------------------------------------------------
// MapHorizFace Class: Map from {(0, 0), (0, 1), (1, 0)} to a horizontal face of a prism
//--------------------------------------------------------------------------------------

// Compute affine mapping coefficients from reference to physical space
// Parameters: coord - face node coordinates
void MapHorizFace::UpdateMap(const real* coord[d_space])
{
  long_real tmp[d_space];

  ref_point=coord[0][2]; 
  map[0]=coord[1][2]-coord[0][2];
  map[1]=coord[2][2]-coord[0][2];
  tmp[0]=map[1]*(coord[1][1]-coord[0][1])-map[0]*(coord[2][1]-coord[0][1]);
  tmp[1]=map[0]*(coord[2][0]-coord[0][0])-map[1]*(coord[1][0]-coord[0][0]);
  tmp[2]=map_2d.GetScale();
  scale=std::sqrt(tmp[0]*tmp[0]+tmp[1]*tmp[1]+tmp[2]*tmp[2]);
}



//--------------------------------------------------------------------------------------
// MapVertFace Class: Map from [0,1]x[0,1] to a vertical face of a prism
//--------------------------------------------------------------------------------------

// Compute affine mapping coefficients from reference square [0,1]x[0,1] to a vertical face
void MapVertFace::UpdateMap(const real* coord[d_space])
{
  map[0]=coord[2][2]-coord[0][2];
  map[1]=coord[1][2]-coord[0][2];
  ref_point=coord[0][2];
  bilin_coef=coord[0][2]-coord[1][2]-coord[2][2]+coord[3][2];
}



//--------------------------------------------------------------------------------------
// MapPrism Class: Map from the reference to a physical prism
//--------------------------------------------------------------------------------------

// Compute affine mapping coefficients from reference to physical space
// Parameters: coord - prism node coordinates
void MapPrism::UpdateMap(const real* coord[d_space])
{
  ref_point=coord[0][2];
  map[0]=coord[1][2]-coord[0][2];
  map[1]=coord[2][2]-coord[0][2];
  map[2]=coord[3][2]-coord[0][2];
  bilin_coef[0]=coord[4][2]-coord[3][2]-coord[1][2]+coord[0][2];
  bilin_coef[1]=coord[5][2]-coord[3][2]-coord[2][2]+coord[0][2];
  vert_scale=std::abs(map[2]+(bilin_coef[0]+bilin_coef[1])/3.);
}

// Get scaling parameters used for physical basis objects
// Parameters: scale - scaling; shift - vector to origin
void MapPrism::GetPhysicalBasisParam(real scale[], real shift[]) const
{
  const real ref[d_space]={0., 0., 0.};

  mapto(ref, shift);
  scale[0]=scale[1]=std::sqrt(map_2d.GetScale());
  scale[2]=vert_scale;
}

