//----------------------------------------------------------------------------------------
// mesh_3d.cpp: Contains definitions for nodes, faces, and prisms in 3D
//----------------------------------------------------------------------------------------

#include <iostream>
#include <vector>
#include "mesh_3d.H"

// Definition of storage vector for vertical split child face id's
std::vector<int> FACE_VERT::child_vert;

//--------------------------------------------------------------------------------------
// NODE_3D Class: Node class in 3D
//--------------------------------------------------------------------------------------

// Default constructor
NODE_3D::NODE_3D(const NODE_2D& node_xy)
{
  id=BAD_INDEX;
  status=0;
  node_2d=node_xy.ID(); 
  scale=0.;
  coord[0]=node_xy.Coord(0); coord[1]=node_xy.Coord(1); coord[2]=0.;
}

// Constructor for hot start input
// Parameters: hot start input file stream
NODE_3D::NODE_3D(std::ifstream& in)
{
  in.read((char*)this, sizeof(*this));
}

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const NODE_3D& node)
{
  out.write((char*)&node, sizeof(node));
  return out;
}



//--------------------------------------------------------------------------------------
// FACE_HORIZ Class: Horizontal face class in 3D
//--------------------------------------------------------------------------------------

// Default constructor
FACE_HORIZ::FACE_HORIZ(const TRIANGLE_2D& face) : mapping(&face.Mapping())
{
  normal[0]=normal[1]=normal[2]=0.; 
  status=n_child=0;
  element_2d=face.ID(); 
  type=id=node[0]=node[1]=node[2]=element[0]=element[1]=parent=child=BAD_INDEX;
}

// Constructor for hot start input
// Parameters: hot start input file stream
FACE_HORIZ::FACE_HORIZ(std::ifstream& in)
{
  in.read((char*)this, sizeof *this);
}

// Compute mapping from the reference face
// Parameters: node_coord - node coordinates
void FACE_HORIZ::ComputeReferenceMapping(const real *node_coord[vert_h])
{
  long_real vector1[d_space]={node_coord[1][0]-node_coord[0][0],
                              node_coord[1][1]-node_coord[0][1],
                              node_coord[1][2]-node_coord[0][2]},
    vector2[d_space]={node_coord[2][0]-node_coord[0][0],
                      node_coord[2][1]-node_coord[0][1],
                      node_coord[2][2]-node_coord[0][2]};

  mapping.UpdateMap(node_coord);
  normal[0]=(vector1[1]*vector2[2]-vector1[2]*vector2[1])/mapping.GetScale();
  normal[1]=(vector1[2]*vector2[0]-vector1[0]*vector2[2])/mapping.GetScale();
  normal[2]=(vector1[0]*vector2[1]-vector1[1]*vector2[0])/mapping.GetScale();
}

// Refine the face using edge bisection
// Parameters: element_2d_child - child triangles in 2D; face_horiz_child - child faces; 
//             node - nodes on the face; node_new - new nodes
void FACE_HORIZ::Refine(TRIANGLE_2D *element_2d_child[4], FACE_HORIZ *face_horiz_child[4], 
                        NODE_3D *node[vert_h], NODE_3D *node_new[vert_h])
{
  const int n_children_new=4;

  // Create new nodes
  for (int i=0; i<vert_h; i++)
    node_new[i]->ZCoordinate()=
      0.5*(node[(i+1)%vert_h]->ZCoordinate()+node[(i+2)%vert_h]->ZCoordinate());
  // Create new horizontal faces
  for (int i=0; i<n_children_new; i++)
  {
    face_horiz_child[i]->Parent()=id;
    face_horiz_child[i]->Type()=type;
    face_horiz_child[i]->Element2D()=element_2d_child[i]->ID();
    face_horiz_child[i]->Element(0)=element[0];
    face_horiz_child[i]->Element(1)=element[1];
    face_horiz_child[i]->Status()=status;
    for (int j=0; j<vert_h; j++)
      for (int k=0; k<vert_h; k++)
      {
        if (element_2d_child[i]->Node(j)==node[k]->Node2D())
        {
          face_horiz_child[i]->Node(j)=node[k]->ID();
          break;
        }
        if (element_2d_child[i]->Node(j)==node_new[k]->Node2D())
        {
          face_horiz_child[i]->Node(j)=node_new[k]->ID();
          break;
        }
      }
  }
  SetChildren(n_children_new, face_horiz_child[0]->ID());
  status|=SPLIT;
}

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const FACE_HORIZ& face)
{
  out.write((char*)&face, sizeof face);
  return out;
}



//--------------------------------------------------------------------------------------
// FACE_VERT Class: Vertical face class in 3D
//--------------------------------------------------------------------------------------

// Default constructor
FACE_VERT::FACE_VERT(const EDGE_2D& edge) : mapping(&edge.Mapping())
{
  normal[0]=edge.Normal(0); normal[1]=edge.Normal(1); normal[2]=0.;
  edge_2d=edge.ID();
  status=n_child_vert=n_child_horiz=0;
  type=edge.Type(); 
  id=node[0]=node[1]=node[2]=node[3]=
    element[0]=element[1]=parent=child_horiz=child_vert_ind=BAD_INDEX;
}

// Constructor for hot start input
// Parameters: hot start input file stream
FACE_VERT::FACE_VERT(std::ifstream& in)
{
  in.read((char*)this, sizeof *this);
}

// Compute mapping from the reference face
// Parameters: node_coord - node coordinates
void FACE_VERT::ComputeReferenceMapping(const real *node_coord[4])
{
  if ((node_coord[0][2]<node_coord[1][2]) || (node_coord[2][2]<node_coord[3][2]) )
  {
    std::cerr << "Degenerate element in the top layer!" << std::endl;
    std::exit(1);
  }
  mapping.UpdateMap(node_coord);
}

// Parameters: n - number of children in vertical split; first - id of the first child face
void FACE_VERT::SetChildrenVert(int n, int first)
{
  assert_debug(n>=0);
  child_vert_ind=child_vert.size();
  n_child_vert=n;
  child_vert.resize(child_vert_ind+n);
  for (int i=0; i<n; i++)
    if (first!=BAD_INDEX)
      child_vert[child_vert_ind+i]=first+i;
    else
      child_vert[child_vert_ind+i]=BAD_INDEX;
}

// Parameters: n - number of children in horizontal split
// Return value: id of the first child face (if present), lvalue
void FACE_VERT::SetChildrenHoriz(int n, int first)
{
  assert_debug(n>=0);
  n_child_horiz=n;
  child_horiz=first;
}

// Refine the face horizontally
// Parameters: edge_2d_child - child edges on the corresponding 2D edge;
//             face_vert_child - child faces; node - nodes on the face; node_new - new node
// Return value: id of the first child face
void FACE_VERT::RefineHoriz(EDGE_2D* edge_2d_child[2], FACE_VERT* face_vert_child[2], 
                            NODE_3D* node[4], NODE_3D* node_new[2])
{
  const int n_children_new=2;

  // Create new nodes
  for (int i=0; i<2; i++)
    node_new[i]->ZCoordinate()=0.5*(node[i]->ZCoordinate()+node[2+i]->ZCoordinate());
  // Create new vertical faces
  for (int i=0; i<n_children_new; i++)
  {
    face_vert_child[i]->Parent()=id;
    face_vert_child[i]->Type()=type;
    face_vert_child[i]->Edge2D()=edge_2d_child[i]->ID();
    face_vert_child[i]->Element(0)=element[0];
    face_vert_child[i]->Element(1)=element[1];
    face_vert_child[i]->Status()=status;
    for (int j=0; j<4; j++)
      face_vert_child[i]->Node(j) = j/2==i ? node[j]->ID() : node_new[j%2]->ID();
  }
  SetChildrenHoriz(n_children_new, face_vert_child[0]->ID());
  status|=HORIZ_SPLIT;
}

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const FACE_VERT& face)
{
  out.write((char*)&face, sizeof face);
  return out;
}



//--------------------------------------------------------------------------------------
// PRISM Class: Element class in 3D
//--------------------------------------------------------------------------------------

// Default constructor
PRISM::PRISM(const TRIANGLE_2D& element) : mapping(&element.Mapping())
{
  element_2d=element.ID();
  status=n_child=0;
  type=id=top=bottom=parent=child=BAD_INDEX;
  for (int i=0; i<vert_h; i++)
    index[i]=element.ElementIndex(i), side[i]=BAD_INDEX;
}

// Constructor for hot start input
// Parameters: hot start input file stream
PRISM::PRISM(std::ifstream& in)
{
  in.read((char*)this, sizeof *this);
}

// Compute mapping from the reference triangle
// Parameters: node_coord - node coordinates
void PRISM::ComputeReferenceMapping(const real *node_coord[2*vert_h])
{
  mapping.UpdateMap(node_coord);
}

// Refine the element
// Parameters: element_2d_child - 2D child elements; edge_2d_new - new 2D edges
//             prism_child - child elements; face_vert_new - new vertical faces;
//             face_vert_child - child vertical faces;
//             face_horiz_child_top - child horizontal faces on the top face;
//             face_horiz_child_bot - child horizontal faces on the bottom face;
//             node_new - new nodes
// Return value: id of the first child element
void PRISM::Refine(TRIANGLE_2D* element_2d_child[4], EDGE_2D* edge_2d_new[vert_h],
                   PRISM* prism_child[4], FACE_VERT* face_vert_new[vert_h],
                   FACE_VERT* face_vert_child[vert_h][2], FACE_HORIZ* face_horiz_child_top[4],
                   FACE_HORIZ* face_horiz_child_bot[4], NODE_3D* node_new[2][vert_h])
{
  const int n_children_new=4;

  // Assign nodes to new vertcal faces
  for (int j=0; j<vert_h; j++)
    for (int k=0; k<2; k++)
      for (int l=0; l<vert_h; l++)
        if (edge_2d_new[j]->Node(k)==node_new[0][l]->Node2D())
          face_vert_new[j]->Node(2*k)=node_new[0][l]->ID(),
          face_vert_new[j]->Node(2*k+1)=node_new[1][l]->ID();

  // Create new elements
  for (int i=0; i<n_children_new; i++)
  {
    prism_child[i]->Element2D()=element_2d_child[i]->ID();
    prism_child[i]->TopFace()=face_horiz_child_top[i]->ID();
    prism_child[i]->BottomFace()=face_horiz_child_bot[i]->ID();
    prism_child[i]->Parent()=id;
    prism_child[i]->Type()=type;
    prism_child[i]->Status()=status;
    face_horiz_child_top[i]->Element(1)=face_horiz_child_bot[i]->Element(0)=prism_child[i]->ID();
    // Update connectivity
    for (int j=0; j<vert_h; j++)
      for (int k=0; k<vert_h; k++)
      {
        if (element_2d_child[i]->Edge(j)==face_vert_new[k]->Edge2D())
        {
          prism_child[i]->SideFace(j)=face_vert_new[k]->ID();
          face_vert_new[k]->Element(prism_child[i]->ElementIndex(j))=prism_child[i]->ID();
          break;
        }
        if (element_2d_child[i]->Edge(j)==face_vert_child[k][0]->Edge2D())
        {
          prism_child[i]->SideFace(j)=face_vert_child[k][0]->ID();
          face_vert_child[k][0]->Element(prism_child[i]->ElementIndex(j))=prism_child[i]->ID();
          break;
        }
        if (element_2d_child[i]->Edge(j)==face_vert_child[k][1]->Edge2D())
        {
          prism_child[i]->SideFace(j)=face_vert_child[k][1]->ID();
          face_vert_child[k][1]->Element(prism_child[i]->ElementIndex(j))=prism_child[i]->ID();
          break;
        }
      }
  }
  SetChildren(n_children_new, prism_child[0]->ID());
  status|=SPLIT;
}

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const PRISM& prism)
{
  out.write((char*)&prism, sizeof prism);
  return out;
}
