//----------------------------------------------------------------------------------------
// quadrature.cpp: Definitions for the class used to store quadrature information
//----------------------------------------------------------------------------------------

#include <cassert>
#include <cmath>
#include "quadrature.H"

extern QUAD_CONTAINER quadrature_index;

//----------------------------------------------------------------------------------------
// QUADRATURE class: Storage for quadrature information
//----------------------------------------------------------------------------------------

// Constructor for non tensor-product quadrature
// Parameters: n_points - number of quadrature points; space - space dimension;
//             p - coordinates of quadrature points; w - quadrature weights
QUADRATURE::QUADRATURE(int n_points, int space, const real p[][d_space], const long_real w[])
{
  n_quad_sub=0;
  n_quad=n_points;
  space_dim=space;
  for (int i=0; i<n_quad; weight[i]=w[i], i++)
    for (int j=0; j<space; point[i][j]=p[i][j], j++);
}

// Constructor for tensor-product quadrature
// Parameters: q1 - main quadrature; q2 - secondary quadrature
QUADRATURE::QUADRATURE(const QUADRATURE *q1, const QUADRATURE *q2)
{
  n_quad_sub=q2->n_quad;
  n_quad=q1->n_quad*q2->n_quad;
  space_dim=q1->space_dim+q2->space_dim;
  assert(space_dim<=d_space);
  for (int i=0, ind=0; i<q1->n_quad; i++)
    for (int j=0; j<q2->n_quad; weight[ind]=q1->weight[i]*q2->weight[j], ind++, j++)
    {
      for (int k=0; k<q1->space_dim; point[ind][k]=q1->point[i][k], k++);
      for (int k=0; k<q2->space_dim; point[ind][q1->space_dim+k]=q2->point[j][k], k++);
    }
}



//----------------------------------------------------------------------------------------
// QUAD_CONTAINER class: Storage for quadrature information for all quadrature objects
//----------------------------------------------------------------------------------------

// Default constructor
QUAD_CONTAINER::QUAD_CONTAINER()
{// Table of Gauss points and weights on [0, 1]
  const real g_point_1[max_order+1][max_quad][d_space]=
  {{{0.5, 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}}, 
  {{0.5*(1.-1./std::sqrt(3.)), 0., 0.}, {0.5*(1.+1./std::sqrt(3.)), 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}}, 
  {{0.5*(1.-std::sqrt(3./5.)), 0., 0.}, {0.5, 0., 0.}, {0.5*(1.+std::sqrt(3./5.)), 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}}};
  const long_real g_weight_1[max_order+1][max_quad]=
  {{1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
  0., 0., 0., 0., 0.}, 
  {0.5, 0.5, 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
  0., 0., 0., 0., 0., 0.}, 
  {5./18.,  4./9., 5./18., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
  0., 0., 0., 0., 0., 0., 0., 0.}};
  // Table of Gauss points and weights on triangle (0, 0), (0, 1), (1, 0)
  const real g_point_2[max_order+1][max_quad][d_space]=
  {{{1./3., 1./3., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}},
  {{0.15505102572168222297, 0.17855872826361643413, 0.}, 
  {0.15505102572168222297, 0.66639024601470131515, 0.}, 
  {0.64494897427831776593, 0.07503111022260812446, 0.}, 
  {0.64494897427831776593, 0.28001991549907406798, 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}},
  {{0.08858795951270398428, 0.10271765480962624650, 0.}, 
  {0.08858795951270398428, 0.45570602024364803562, 0.},
  {0.08858795951270398428, 0.80869438567766971371, 0.}, 
  {0.40946686444073471245, 0.06655406783916449631, 0.}, 
  {0.40946686444073471245, 0.29526656777963267153, 0.}, 
  {0.40946686444073471245, 0.52397906772010083287, 0.},
  {0.78765946176084700170, 0.02393113228708061702, 0.}, 
  {0.78765946176084700170, 0.10617026911957647139, 0.},
  {0.78765946176084700170, 0.18840940595207233965, 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.},
  {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}, {0., 0., 0.}}};
  const long_real g_weight_2[max_order+1][max_quad]=
  {{0.5, 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
  0., 0., 0., 0., 0., 0.},
  {0.15902069087198861252, 0.15902069087198861252, 
  0.09097930912801141523, 0.09097930912801141523, 0., 0., 0., 0., 0., 0., 0., 0., 0.,
  0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.},
  {0.05581442048304434411, 0.08930307277287098666, 0.05581442048304434411, 
  0.06367808509988504007, 0.10188493615981610019, 0.06367808509988504007, 
  0.01939638330595946231, 0.03103421328953515426, 0.01939638330595946231, 0., 0., 0.,
  0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.}};
  
  //Generator: [ Fully symmetric ] 3 points degree 2
 //( 0.166666666666666666666666666666666, 0.166666666666666666666666666666666, ) 
 //Corresponding weight: 
 //0.166666666666666666666666666666666,
  //
 //  Generator: [ Fully symmetric ] 6 points degree 3
 //( 0.109039009072877212324834667569910, 0.231933368553030572496784561174692, ) 
 //Corresponding weight: 
 //0.0833333333333333333333333333333333,
  
//  Generator: [ Fully symmetric ]  6 points degree 4
//( 0.0915762135097707434595714634022015, 0.0915762135097707434595714634022015, ) 
//Corresponding weight: 
//0.0549758718276609338191631624501052,
//Generator: [ Fully symmetric ] 
//( 0.445948490915964886318329253883051, 0.445948490915964886318329253883051, ) 
//Corresponding weight: 
//0.111690794839005732847503504216561,

// Generator: [ Fully symmetric ]   7 points degree 5
// ( 0.333333333333333333333333333333333, 0.333333333333333333333333333333333, ) 
// Corresponding weight: 
// 0.1125,
// Generator: [ Fully symmetric ] 
// ( 0.101286507323456338800987361915123, 0.101286507323456338800987361915123, ) 
// Corresponding weight: 
// 0.0629695902724135762978419727500906,

// Generator: [ Fully symmetric ] 
// ( 0.470142064105115089770441209513447, 0.470142064105115089770441209513447, ) 
// Corresponding weight: 
// 0.0661970763942530903688246939165759,

  // Node coordinates on the reference triangle
  const real node_2[vert_h][d_space]={{0., 0., 0.}, {1., 0., 0.}, {0., 1., 0.}};
  // Node coordinates on the reference prism
  const real node_3[2*vert_h][d_space]=
  {{0., 0., 0.}, {1., 0., 0.}, {0., 1., 0.}, {0., 0., 1.}, {1., 0., 1.}, {0., 1., 1.}};

  // Coordinates for the vertices of {{0., 0.}, {1., 0.}, {0., 1.}} triangle
  QUADRATURE triangle_node_quad_obj(vert_h, 2, node_2, g_weight_2[0]);
  triangle_node_quad=storage.size();
  storage.push_back(triangle_node_quad_obj);

  // Coordinates for the vertices of the reference prism
  QUADRATURE prism_node_quad_obj(2*vert_h, d_space, node_3, g_weight_2[0]);
  prism_node_quad=storage.size();
  storage.push_back(prism_node_quad_obj);

  for (int i=0; i<=max_order; i++)
  {
    // Quadrature on [0,1] segment
    QUADRATURE edge_quad_obj(i+1, 1, g_point_1[i], g_weight_1[i]);
    edge_quad[i]=storage.size();
    storage.push_back(edge_quad_obj);

    // Quadrature on {{0., 0.}, {1., 0.}, {0., 1.}} triangle
    QUADRATURE triangle_quad_obj((i+1)*(i+1), 2, g_point_2[i], g_weight_2[i]);
    triangle_quad[i]=storage.size();
    storage.push_back(triangle_quad_obj);

    // Quadrature on {{0., 0., 0.}, {1., 0., 0.}, {0., 1., 0.}, {0., 0., 1.}, 
    // {1., 0., 1.}, {0., 1., 1.}} prism
    QUADRATURE prism_quad_obj(GetTriangleQuad(i), GetEdgeQuad(i));
    prism_quad[i]=storage.size();
    storage.push_back(prism_quad_obj);

    // Quadrature on the unit square
    QUADRATURE vert_face_quad_obj(GetEdgeQuad(i), GetEdgeQuad(i));
    vert_face_quad[i]=storage.size();
    storage.push_back(vert_face_quad_obj);
  }
}
