//----------------------------------------------------------------------------------------
// Analytical solution for test problems
//----------------------------------------------------------------------------------------

#include <cmath>
#include "pde.H"
#include "test.H"

extern PROBLEM_DESCRIPTION problem_info;

const real amp=0.01;

// True solution for the test problem
void solution_analytic(real t, const real coord[d_space], real func[d_var])
{
  real coef1=0.1*problem_info.pi, coef2=0.5*problem_info.pi;

  func[U]=amp/coef1*std::sin(coef1*coord[0])*std::sin(coef2*(coord[2]+1.));
  func[V]=0.;
  func[W]=amp/coef2*std::cos(coef1*coord[0])*std::cos(coef2*(coord[2]+1.));
#ifdef BAROCLINIC_KEY
  func[T]=20.;
  func[S]=35.;
  func[RO]=density(problem_info, coord, func);
#endif
}

// Gradient of the true solution
void solution_analytic_grad(real t, const real coord[d_space], real grad[d_var][d_space])
{
  real coef1=0.1*problem_info.pi, coef2=0.5*problem_info.pi;

  grad[U][0]=amp*std::cos(coef1*coord[0])*std::sin(coef2*(coord[2]+1.));
  grad[U][1]=0.;
  grad[U][2]=amp*coef2/coef1*std::sin(coef1*coord[0])*std::cos(coef2*(coord[2]+1.));
  grad[V][0]=0.;
  grad[V][1]=0.;
  grad[V][2]=0.;
  grad[W][0]=-amp*coef1/coef2*std::sin(coef1*coord[0])*std::cos(coef2*(coord[2]+1.));
  grad[W][1]=0.;
  grad[W][2]=-amp*std::cos(coef1*coord[0])*std::sin(coef2*(coord[2]+1.));
#ifdef BAROCLINIC_KEY
  grad[T][0]=0.;
  grad[T][1]=0.;
  grad[T][2]=0.;
  grad[S][0]=0.;
  grad[S][1]=0.;
  grad[S][2]=0.;
#endif
}

// Laplacian of the true solution
void solution_analytic_laplacian(real t, const real coord[d_space], real lapl[d_var][d_space])
{
  real coef1=0.1*problem_info.pi, coef2=0.5*problem_info.pi;

  lapl[U][0]=-amp*coef1*std::sin(coef1*coord[0])*std::sin(coef2*(coord[2]+1.));
  lapl[U][1]=0.;
  lapl[U][2]=-amp*coef2*coef2/coef1*std::sin(coef1*coord[0])*std::sin(coef2*(coord[2]+1.));
  lapl[V][0]=0.;
  lapl[V][1]=0.;
  lapl[V][2]=0.;
#ifdef BAROCLINIC_KEY
  lapl[T][0]=0.;
  lapl[T][1]=0.;
  lapl[T][2]=0.;
  lapl[S][0]=0.;
  lapl[S][1]=0.;
  lapl[S][2]=0.;
#endif
}

// Time derivative of the true solution for the test problem
void solution_analytic_dt(real t, const real coord[d_space], real t_deriv[d_var])
{
  t_deriv[U]=0.;
  t_deriv[V]=0.;
#ifdef BAROCLINIC_KEY
  t_deriv[T]=0.;
  t_deriv[S]=0.;
#endif
}

// Analytic solution for the elevation
void solution_analytic_elevation(real t, const real coord[2], real result[])
{
  result[0]=amp*coord[0];
}

// Analytic solution for the elevation gradient
void solution_analytic_elevation_grad(real t, const real coord[2], real elev_grad[2])
{
  elev_grad[0]=amp;
  elev_grad[1]=0.;
}

// Time derivative of the true solution for the elevation
real solution_analytic_elevation_dt(real t, const real coord[d_space], real result[])
{
  result[0]=0.0;
}

#ifdef BAROCLINIC_KEY
// Equation of state for density
long_real density(const real coord[d_space], const real var[d_var])
{
  real coef=0.2*problem_info.pi;
  return amp*amp/coef/coef/problem_info.g*std::cos(coef*coord[0]);
}

// Forcing function accounting for density terms
void density_forcing(real t, real xi, const real xi_grad[d_space_2d], const real coord[d_space], real result[])
{
  real coord_top[d_space]={coord[0], coord[1], xi}, var[d_var], ro=density(coord_top, var);
  real coef=0.2*problem_info.pi;

  result[0]=-amp*amp/coef*(xi-coord[2])*std::sin(coef*coord[0])+problem_info.g*ro*xi_grad[0];
  result[1]=problem_info.g*ro*xi_grad[1];
}
#endif
